/*
 * RFBFramebuffer.m
 * A simple off-screen framebuffer optimized for drawing rectangles and raw images.
 *
 * Copyright (c) 1997 21st Century Software, New York City.
 *
 * See the file Reference/License.rtf for information on usage and
 * redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by David Young <dwy@ace.net>, 1997.
 */

#import <AppKit/AppKit.h>

#import "RFB.h"

unsigned RFBConvertPixel8ToPixel32(unsigned char pixel8)
{
    unsigned long pixel32 = 0;

    if (NSHostByteOrder() == NS_LittleEndian) {
        pixel32 += ((pixel8 >> 6) & 3) * 0x400000;
        pixel32 += ((pixel8 >> 4) & 3) * 0x004000;
        pixel32 += ((pixel8 >> 2) & 3) * 0x000040;
    }
    else {
        pixel32 += ((pixel8 >> 6) & 3) * 0x40000000;
        pixel32 += ((pixel8 >> 4) & 3) * 0x00400000;
        pixel32 += ((pixel8 >> 2) & 3) * 0x00004000;
    }        

    return pixel32;
}

unsigned RFBConvertPixel16ToPixel32(unsigned short pixel16)
{
    unsigned long pixel32 = 0;

    if (NSHostByteOrder() == NS_LittleEndian) {
        pixel32 += ((pixel16 >> 12) & 15) * 0x100000;
        pixel32 += ((pixel16 >>  8) & 15) * 0x001000;
        pixel32 += ((pixel16 >>  4) & 15) * 0x000010;
    }
    else {
        pixel32 += ((pixel16 >> 12) & 15) * 0x10000000;
        pixel32 += ((pixel16 >>  8) & 15) * 0x00100000;
        pixel32 += ((pixel16 >>  4) & 15) * 0x00001000;
    }

   return pixel32;
}

@implementation RFBFramebuffer

- (id)initWithSize:(NSSize)size
{
    if (self = [super init]) {
        _width = size.width;
        _height = size.height;
        _bytesPerRow = size.width * sizeof(unsigned);
        _minimumUpdateRect = NSMakeRect(0, 0, 0, 0);
        _framebufferCover = [[NSMutableData alloc] initWithLength:_bytesPerRow*_width];
        _framebufferPtr = [_framebufferCover mutableBytes];
    }

    return self;
}

- (NSString *)description
{
    return [NSString stringWithFormat:@"<%@[0x%lx] %dx%d bytesPerRow=%d minUpdateRect=%@ framebufferPtr=0x%lx>",
        [[self class] description], self,
        _width, _height, _bytesPerRow, NSStringFromRect(_minimumUpdateRect),
        _framebufferPtr];
}

- (void)dealloc
{
    [_framebufferCover release];

    [super dealloc];
}

- (NSImage *)framebufferImage
{
    NSImage *image = [[[NSImage alloc] initWithSize:NSMakeSize(_width, _height)] autorelease];
    const unsigned char *pixelPointers[5];
    NSBitmapImageRep *imageRep;
    
    pixelPointers[0] = (const unsigned char *)_framebufferPtr;
    imageRep = [[[NSBitmapImageRep alloc]
        initWithBitmapDataPlanes:pixelPointers
                      pixelsWide:_width
                      pixelsHigh:_height
                   bitsPerSample:8
                 samplesPerPixel:3
                        hasAlpha:NO
                        isPlanar:NO
                  colorSpaceName:NSCalibratedRGBColorSpace
                     bytesPerRow:_bytesPerRow
                    bitsPerPixel:32] autorelease];

    [image addRepresentation:imageRep];

    return image;
}

- (NSData *)subPixelsInRect:(NSRect)rect
{
    unsigned bytesPerRow = rect.size.width * sizeof(unsigned);
    NSMutableData *subData = [NSMutableData dataWithLength:bytesPerRow*rect.size.height];
    unsigned *subDataPtr = [subData mutableBytes];
    unsigned x, y;

    for (y = rect.origin.y; y < rect.origin.y+rect.size.height; ++y)
        for (x = rect.origin.x; x < rect.origin.x+rect.size.width; ++x)
            *subDataPtr++ = _framebufferPtr[(y*_width)+x];

    return subData;
}

- (void)compositePixelData:(NSData *)pixelData targetRect:(NSRect)target bitsPerPixel:(unsigned)bitsPerPixel
{
    unsigned x, y;

    switch(bitsPerPixel) {
        case 32: {
            const unsigned *pixelPtr = [pixelData bytes];                          
            for (y = target.origin.y; y < (target.origin.y + target.size.height); ++y)
                for (x = target.origin.x; x < (target.origin.x + target.size.width); ++x)
                    _framebufferPtr[(y * _width) + x] = NSSwapBigIntToHost(*pixelPtr++);
            break;
        }

        case 16: {
            const unsigned short *pixelPtr = [pixelData bytes];                          
            for (y = target.origin.y; y < (target.origin.y + target.size.height); ++y)
                for (x = target.origin.x; x < (target.origin.x + target.size.width); ++x)
                    _framebufferPtr[(y * _width) + x] = \
                        RFBConvertPixel16ToPixel32(NSSwapBigShortToHost(*pixelPtr++));
            break;
        }

        case 8: {
            const unsigned char *pixelPtr = [pixelData bytes];                          
            for (y = target.origin.y; y < (target.origin.y + target.size.height); ++y) 
                for (x = target.origin.x; x < (target.origin.x + target.size.width); ++x)
                    _framebufferPtr[(y * _width) + x] = \
                        RFBConvertPixel8ToPixel32(*pixelPtr++);
            break;
        }

        default:
            [NSException raise:NSInvalidArgumentException
                        format:@"%@ invalid bit depth %d",
                NSStringFromSelector(_cmd), bitsPerPixel];
    }

    _minimumUpdateRect = NSUnionRect(_minimumUpdateRect, target);
}

- (void)copyPixelsFromRect:(NSRect)srcRect toRect:(NSRect)destRect
{
    [self compositePixelData:[self subPixelsInRect:srcRect]
                  targetRect:destRect
                bitsPerPixel:32];
}

- (void)fillRect:(NSRect)rect color:(unsigned)color
{
    unsigned x, y;

    for (y = rect.origin.y; y < (rect.origin.y + rect.size.height); ++y)
        for (x = rect.origin.x; x < (rect.origin.x + rect.size.width); ++x)
            _framebufferPtr[(y * _width) + x] = color;
            
    _minimumUpdateRect = NSUnionRect(_minimumUpdateRect, rect);
}

- (void)fillRectList:(const NSRect *)rectList colorList:(const unsigned *)colorList count:(unsigned)count
{
    unsigned i = 0;

    while (i < count) {
        NSRect rect = rectList[i];
        unsigned x, y;
        unsigned color = colorList[i];

        for (y = rect.origin.y; y < (rect.origin.y + rect.size.height); ++y)
            for (x = rect.origin.x; x < (rect.origin.x + rect.size.width); ++x)
                _framebufferPtr[(y * _width) + x] = color;

        _minimumUpdateRect = NSUnionRect(_minimumUpdateRect, rect);
        ++i;
    }
}

- (void)updateRect:(NSRect)updateRect
{
    const unsigned char *pixelPointers[5];
    NSBitmapImageRep *imageRep;

    if (updateRect.size.width == _width && updateRect.size.height == _height)
        pixelPointers[0] = (const unsigned char *)_framebufferPtr;
    else
        pixelPointers[0] = [[self subPixelsInRect:updateRect] bytes];

    imageRep = [[[NSBitmapImageRep alloc]
        initWithBitmapDataPlanes:pixelPointers
                      pixelsWide:updateRect.size.width
                      pixelsHigh:updateRect.size.height
                   bitsPerSample:8
                 samplesPerPixel:3
                        hasAlpha:NO
                        isPlanar:NO
                  colorSpaceName:NSCalibratedRGBColorSpace
                     bytesPerRow:updateRect.size.width*sizeof(unsigned)
                    bitsPerPixel:32] autorelease];

    // correct for coordinate weirdness
    updateRect.origin.y = (_height/2)+((_height/2)-updateRect.origin.y);
    updateRect.origin.y -= updateRect.size.height;
    
    [imageRep drawAtPoint:updateRect.origin];
}

- (void)update
{
    [self updateRect:NSMakeRect(0, 0, _width, _height)];
    _minimumUpdateRect = NSMakeRect(0, 0, 0, 0);
}

- (void)updateMinimum
{
    [self updateRect:_minimumUpdateRect];
    _minimumUpdateRect = NSMakeRect(0, 0, 0, 0);
}

@end

