/*
 * RFBHextileSubrectangle.m
 *
 *
 * Copyright (c) 1997 21st Century Software, New York City.
 *
 * See the file Reference/License.rtf for information on usage and
 * redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by David Young <dwy@ace.net>, 1997.
 */

#import "RFB.h"

@implementation RFBHextileSubrectangle

+ (RFBHextileSubrectangle *)subrectangleWithLocation:(NSRect)location superrectangle:(RFBHextileRectangle *)superrect adaptor:(RFBProtocolAdaptor *)adaptor
{
    return [[[self alloc] initWithLocation:location
                            superrectangle:superrect
                                   adaptor:adaptor] autorelease];
}

- (id)initWithLocation:(NSRect)location superrectangle:(RFBHextileRectangle *)superrect adaptor:(RFBProtocolAdaptor *)adaptor
{
    if (self = [super init]) {
        _location = location;
        _subEncodingMask = [[adaptor socket] readByte];

        if (_subEncodingMask & RFBHextileRaw) {
            unsigned i;

            _bitsPerPixel = [adaptor bitsPerPixel];
            i = location.size.width*location.size.height*(_bitsPerPixel/8);
            _pixelData = [[[adaptor socket] readDataOfLength:i] retain];
        }
        else {
            if (_subEncodingMask & RFBHextileBackgroundSpecified) {
                _backgroundColor = [adaptor read24BitColor];
                [superrect setLastBackgroundColor:_backgroundColor];
            }
            else
                _backgroundColor = [superrect lastBackgroundColor];

            if (_subEncodingMask & RFBHextileForegroundSpecified)
                _foregroundColor = [adaptor read24BitColor];

            if (_subEncodingMask & RFBHextileAnySubrects) {
                NSRect *rectList = NULL;
                unsigned *colorList = NULL;
                octet i;

                _numberOfRects = [[adaptor socket] readByte];                
                _rectListCover = [[NSMutableData alloc] initWithLength:_numberOfRects*sizeof(NSRect)];
                rectList = [_rectListCover mutableBytes];

                if (_subEncodingMask & RFBHextileSubrectsColoured) {
                    _colorListCover = [[NSMutableData alloc] initWithLength:_numberOfRects*sizeof(unsigned)];
                    colorList = [_colorListCover mutableBytes];
                }
                
                for (i = 0; i < _numberOfRects; ++i) {
                    octet xy, hw;
                    if (_subEncodingMask & RFBHextileSubrectsColoured)
                        colorList[i] = [adaptor read24BitColor];

                    xy = [[adaptor socket] readByte];
                    hw = [[adaptor socket] readByte];

                    rectList[i].origin.x = _location.origin.x + (xy >> 4);
                    rectList[i].origin.y = _location.origin.y + (xy & 0xf);
                    rectList[i].size.width = ((hw >> 4)+1);
                    rectList[i].size.height = ((hw & 0xF)+1);
                }
            }
        }
    }

    return self;
}

- (void)dealloc
{
    [_pixelData release];
    [_rectListCover release];
    [_colorListCover release];

    [super dealloc];
}

- (NSString *)description
{
    return [NSString stringWithFormat:@"<%@[0x%lx] location=%@ subEncodingMask=0x%02x>",
        [[self class] description], self,
        NSStringFromRect(_location), (int)_subEncodingMask];
}

- (void)drawWithFramebuffer:(RFBFramebuffer *)framebuffer
{
    if (_subEncodingMask & RFBHextileRaw) {
        [framebuffer compositePixelData:_pixelData
                             targetRect:_location
                           bitsPerPixel:_bitsPerPixel];
    }
    else {
        [framebuffer fillRect:_location color:_backgroundColor];

        if (_subEncodingMask & RFBHextileAnySubrects) {
            const NSRect *rectList = [_rectListCover bytes];
            const unsigned *colorList = [_colorListCover bytes];
            octet i;

            if (_subEncodingMask & RFBHextileSubrectsColoured)
                [framebuffer fillRectList:rectList colorList:colorList count:_numberOfRects];
            else 
                for (i = 0; i < _numberOfRects; ++i)
                    [framebuffer fillRect:rectList[i] color:_foregroundColor];
        }
    }
}

- (id)initWithCoder:(NSCoder *)coder
{
    if (self = [super init]) {
        _location = [coder decodeRect];
        [coder decodeValueOfObjCType:@encode(unsigned) at:&_subEncodingMask];
        [coder decodeValueOfObjCType:@encode(unsigned) at:&_bitsPerPixel];
        _pixelData = [[coder decodeDataObject] retain];
        [coder decodeValueOfObjCType:@encode(unsigned) at:&_backgroundColor];
        [coder decodeValueOfObjCType:@encode(unsigned) at:&_foregroundColor];
        [coder decodeValueOfObjCType:@encode(octet) at:&_numberOfRects];
        _rectListCover = [[coder decodeDataObject] retain];
        _colorListCover = [[coder decodeDataObject] retain];
    }

    return self;
}

- (void)encodeWithCoder:(NSCoder *)coder
{
    [coder encodeRect:_location];
    [coder encodeValueOfObjCType:@encode(unsigned) at:&_subEncodingMask];
    [coder encodeValueOfObjCType:@encode(unsigned) at:&_bitsPerPixel];
    [coder encodeDataObject:_pixelData];
    [coder encodeValueOfObjCType:@encode(unsigned) at:&_backgroundColor];
    [coder encodeValueOfObjCType:@encode(unsigned) at:&_foregroundColor];
    [coder encodeValueOfObjCType:@encode(octet) at:&_numberOfRects];
    [coder encodeDataObject:_rectListCover];
    [coder encodeDataObject:_colorListCover];
}

@end
