/*
 * RFBMessage.m
 *
 *
 * Copyright (c) 1997 21st Century Software, New York City.
 *
 * See the file Reference/License.rtf for information on usage and
 * redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by David Young <dwy@ace.net>, 1997.
 */

#import "RFB.h"

@implementation RFBMessage

+ (RFBMessage *)messageWithAdaptor:(RFBProtocolAdaptor *)adaptor
{
    TCStream *stream = [adaptor socket];
    octet type = [stream readByte];

    if (adaptor == nil)
        return nil;

    switch(type) {
        case RFBFramebufferUpdate: {
            NSMutableArray *rectangles = [NSMutableArray array];
            unsigned short rectCount = 0, rectCtr = 0;

            [stream readByte];			// padding
            rectCount = [stream readUnsignedShort];
            for (rectCtr = 0; rectCtr < rectCount; ++rectCtr)
                [rectangles addObject:[RFBRectangle rectangleWithAdaptor:adaptor]];

            return [self framebufferUpdateMessageWithRectangles:rectangles];
        }            

        case RFBServerCutText: {
            unsigned length;
            NSString *text;
            
            [stream readByte];			// padding
            length = [stream readUnsignedInt];
            text = [NSString stringWithData:[stream readDataOfLength:length]
                                   encoding:NSASCIIStringEncoding];
                
            return [self cutTextMessageWithText:text];
        }

        // ignored, but recognized
        case RFBSetColourMapEntries: {
            unsigned short colourCount = 0, i = 0;;
            
            [stream readByte];			// padding
            [stream readUnsignedShort];	// first color
            colourCount = [stream readUnsignedShort];
            for (i = 0; i < colourCount; ++i) {
                [stream readUnsignedShort];	// red
                [stream readUnsignedShort];	// green
                [stream readUnsignedShort];	// blue
            }
                
            return [[[self alloc] initWithType:RFBSetColourMapEntries] autorelease];
        }

        case RFBBell:
            return [self bellMessage];

        default:
            [NSException raise:NSInvalidArgumentException
                        format:@"%@: unknown message type %d", NSStringFromSelector(_cmd), type];
    }

    return nil; // never reached
}

+ (RFBMessage *)framebufferUpdateMessageWithRectangles:(NSArray *)rectangles;
{
    return [[[RFBFramebufferUpdateMessage alloc] initWithRectangles:rectangles] autorelease];
}

+ (RFBMessage *)bellMessage
{
    return [[[self alloc] initWithType:RFBBell] autorelease];
}

+ (RFBMessage *)cutTextMessageWithText:(NSString *)text
{
    return [[[RFBCutTextMessage alloc] initWithText:text] autorelease];
}

- (id)initWithType:(octet)type
{
    if (self = [super init]) {
        _type = type;
    }

    return self;
}

- (octet)type
{
    return _type;
}

- (NSString *)description
{
    return [NSString stringWithFormat:@"<%@[0x%lx] type=%d>",
        [[self class] description], self,
        _type];
}

- (id)copyWithZone:(NSZone *)zone
{
    return [self retain];
}

- (id)initWithCoder:(NSCoder *)coder
{
    if (self = [super init]) {
        [coder decodeValueOfObjCType:@encode(octet) at:&_type];
    }

    return self;
}

- (void)encodeWithCoder:(NSCoder *)coder
{
    [coder encodeValueOfObjCType:@encode(octet) at:&_type];
}

@end

@implementation RFBFramebufferUpdateMessage

- (id)initWithRectangles:(NSArray *)rectangles
{
    if (self = [super initWithType:RFBFramebufferUpdate]) {
        _rectangles = [rectangles retain];
    }

    return self;
}

- (void)dealloc
{
    [_rectangles release];

    [super dealloc];
}

- (NSString *)description
{
    return [NSString stringWithFormat:@"<%@[0x%lx] rectangles=%@>",
        [[self class] description], self,
        [_rectangles description]];
}

- (NSArray *)rectangles
{
    return _rectangles;
}

- (id)initWithCoder:(NSCoder *)coder
{
    if (self = [super initWithCoder:coder]) {
        _rectangles = [[coder decodeObject] retain];
    }

    return self;
}

- (void)encodeWithCoder:(NSCoder *)coder
{
    [super encodeWithCoder:coder];
    
    [coder encodeObject:_rectangles];
}

@end

@implementation RFBCutTextMessage

- (id)initWithText:(NSString *)text
{
    if (self = [super initWithType:RFBServerCutText]) {
        _text = [text retain];
    }

    return self;
}

- (void)dealloc
{
    [_text release];

    [super dealloc];
}

- (NSString *)description
{
    return [NSString stringWithFormat:@"<%@[0x%lx] text=\"%@\">",
        [[self class] description], self,
        _text];
}

- (NSString *)text
{
    return _text;
}

- (id)initWithCoder:(NSCoder *)coder
{
    if (self = [super initWithCoder:coder]) {
        _text = [[coder decodeObject] retain];
    }

    return self;
}

- (void)encodeWithCoder:(NSCoder *)coder
{
    [super encodeWithCoder:coder];

    [coder encodeObject:_text];
}

@end
