/*
 * RFBProtocolAdaptor.m
 *
 *
 * Copyright (c) 1997 21st Century Software, New York City.
 *
 * See the file Reference/License.rtf for information on usage and
 * redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by David Young <dwy@ace.net>, 1997.
 */

#import "RFB.h"
#import "rfbauth.h"

NSString *RFBProtocolVersion = @"RFB 003.003\n";
NSString *RFBProtocolException = @"RFBProtocolException";

@implementation RFBProtocolAdaptor

+ (RFBProtocolAdaptor *)adaptorWithHost:(NSString *)host port:(int)port
{
    return [[[self alloc] initWithHost:host port:port] autorelease];
}

- (id)initWithHost:(NSString *)host port:(int)port
{
    if (self = [super init]) {
        if (port < 100)
            port += RFBServerPort;

        _socket = [[TCSocketConnector defaultConnector] socketConnectedToHostName:host
                                                                             port:port];
        [_socket setEndianness:NS_BigEndian];
        [_socket retain];
        
        _host = [host copy];
        _port = port;
    }

    return self;
}

- (TCSocket *)socket
{
    return _socket;
}

- (unsigned short)width
{
    return _width;
}

- (unsigned short)height;
{
    return _height;
}

- (octet)bitsPerPixel;
{
    return _bitsPerPixel;
}

- (octet)depth;
{
    return _depth;
}

- (BOOL)bigEndian;
{
    return _bigEndian;
}

- (BOOL)trueColour;
{
    return _trueColour;
}

- (unsigned short)redMax;
{
    return _redMax;
}

- (unsigned short)greenMax;
{
    return _greenMax;
}

- (unsigned short)blueMax;
{
    return _blueMax;
}

- (octet)redShift;
{
    return _redShift;
}

- (octet)greenShift;
{
    return _greenShift;
}

- (octet)blueShift;
{
    return _blueShift;
}

- (NSString *)desktopName;
{
    return _desktopName;
}

- (int)versionMajor
{
    return _versionMajor;
}

- (int)versionMinor
{
    return _versionMinor;
}

- (NSString *)readServerVersionMessage
{
    NSData *data;
    NSString *string;

    data = [_socket readDataOfLength:12];
    string = [NSString stringWithData:data encoding:NSASCIIStringEncoding];

    if ([string hasPrefix:@"RFB"]) {
        NSScanner *scanner = [NSScanner scannerWithString:string];

        if ([scanner scanString:@"RFB " intoString:NULL] &&
            [scanner scanInt:&_versionMajor] &&
            [scanner scanString:@"." intoString:NULL] &&
            [scanner scanInt:&_versionMinor])
            return string;
    }

    [NSException raise:RFBProtocolException format:@"%@ is not an RFB server", _host];
    return nil;		/* never reached */
}

- (void)sendClientVersionMessage
{
    [_socket writeString:RFBProtocolVersion];
}

- (int)readServerAuthScheme
{
    int scheme = [_socket readUnsignedInt];

    switch(scheme) {
        case RFBConnectionFailed: {
            int reasonLength = [_socket readUnsignedInt];
            NSString *reason = [NSString stringWithData:[_socket readDataOfLength:reasonLength] encoding:NSASCIIStringEncoding];
            
            [NSException raise:RFBProtocolException format:@"%@", reason];
        }

        case RFBNoAuthorization:
        case RFBVNCAuthorization:
            return scheme;

        default:
            [NSException raise:RFBProtocolException
                        format:@"Unknown authentication scheme from RFB %d", scheme];
    }

    return -1; /* never reached */
}

- (NSData *)readServerAuthChallenge
{
    return [_socket readDataOfLength:16];
}

- (BOOL)authorizeWithPassword:(NSString *)passString challenge:(NSData *)challengeData
{
    char *challenge = [[[challengeData mutableCopy] autorelease] mutableBytes];
    char *password;
    int result;

    if (passString == nil)
        return NO;
    
    password = [[NSMutableData dataWithLength:[passString cStringLength]] mutableBytes];
    [passString getCString:password];
    
    // truncate and encrypt password
    if (strlen(password) > 8)
        password[8] = '\0';
    vncEncryptBytes(challenge, password);

    // send encrypted password to server
    [_socket writeDataOfLength:CHALLENGESIZE at:challenge];

    // return result code
    result = [_socket readUnsignedInt];
    switch (result) {
        case RFBAuthorizationOK:		return YES;
        case RFBAuthorizationFailed:		return NO;
        case RFBAuthorizationTooMany:
            [NSException raise:RFBProtocolException
                        format:@"VNC authentication failed - too many tries"];
            break;
        default:
            [NSException raise:RFBProtocolException
                        format:@"Unknown VNC authentication result: %d", result];
            break;
    }

    return -1; /* never reached */
}

- (void)sendClientInitializationWithSharedDesktop:(BOOL)sharedFlag
{
    [_socket writeByte:sharedFlag];
}

- (void)readServerInitializationMessage
{
    int nameLength;

    _width = [_socket readUnsignedShort];
    _height = [_socket readUnsignedShort];
    _bitsPerPixel = [_socket readByte];
    _depth = [_socket readByte];
    _bigEndian = ([_socket readByte] != 0);
    _trueColour = ([_socket readByte] != 0);
    _redMax = [_socket readUnsignedShort];
    _greenMax = [_socket readUnsignedShort];
    _blueMax = [_socket readUnsignedShort];
    _redShift = [_socket readByte];
    _greenShift = [_socket readByte];
    _blueShift = [_socket readByte];

    // padding
    [_socket readDataOfLength:3];

    nameLength = [_socket readUnsignedInt];
    _desktopName = [NSString stringWithData:[_socket readDataOfLength:nameLength]
                                   encoding:NSASCIIStringEncoding];
}

- (void)sendEncodings:(NSArray *)encodings;
{
    NSEnumerator *encodingEnumerator = [encodings reverseObjectEnumerator];
    NSNumber *encoding;

    [_socket writeByte:RFBSetEncodings];
    [_socket writeByte:0];
    [_socket writeUnsignedShort:[encodings count]];

    while (encoding = [encodingEnumerator nextObject])
        [_socket writeUnsignedInt:[encoding unsignedIntValue]];
}

- (void)sendBitsPerPixel:(octet)bitsPerPixel depth:(octet)depth bigEndian:(octet)bigEndian trueColour:(octet)trueColour redMax:(unsigned short)redMax greenMax:(unsigned short)greenMax blueMax:(unsigned short)blueMax redShift:(octet)redShift greenShift:(octet)greenShift blueShift:(octet)blueShift
{
    [_socket writeByte:RFBSetPixelFormat];
    [_socket writeByte:0];
    [_socket writeByte:0];
    [_socket writeByte:0];
    [_socket writeByte:bitsPerPixel];
    [_socket writeByte:depth];
    [_socket writeByte:bigEndian];
    [_socket writeByte:trueColour];
    [_socket writeUnsignedShort:redMax];
    [_socket writeUnsignedShort:greenMax];
    [_socket writeUnsignedShort:blueMax];
    [_socket writeByte:redShift];
    [_socket writeByte:greenShift];
    [_socket writeByte:blueShift];
    [_socket writeByte:0];
    [_socket writeByte:0];
    [_socket writeByte:0];

    _bitsPerPixel = bitsPerPixel;
    _depth = depth;
    _bigEndian = bigEndian;
    _trueColour = trueColour;
    _redMax = redMax;
    _greenMax = greenMax;
    _blueMax = blueMax;
    _redShift = redShift;
    _greenShift = greenShift;
    _blueShift = blueShift;
}

- (void)sendFramebufferUpdateRequestForRect:(NSRect)r incremental:(BOOL)incremental
{
    [_socket writeByte:RFBFramebufferUpdateRequest];
    [_socket writeByte:incremental];
    [_socket writeUnsignedShort:(unsigned short)r.origin.x];
    [_socket writeUnsignedShort:(unsigned short)r.origin.y];
    [_socket writeUnsignedShort:(unsigned short)r.size.width];
    [_socket writeUnsignedShort:(unsigned short)r.size.height];
}

- (void)sendPointerEventWithButtonMask:(octet)buttonMask location:(NSPoint)location
{
    [_socket writeByte:RFBPointerEvent];
    [_socket writeByte:buttonMask];
    [_socket writeUnsignedShort:(unsigned short)location.x];
    [_socket writeUnsignedShort:(unsigned short)location.y];
}

- (void)sendKeyEventWithPressedFlag:(BOOL)isPressed key:(unsigned)key
{
    [_socket writeByte:RFBKeyEvent];
    [_socket writeByte:isPressed];
    [_socket writeUnsignedShort:0];	// padding
    [_socket writeUnsignedInt:key];
}

- (void)abort
{
    [_socket closeStream];
}

- (void)dealloc
{
    [_socket release];
    [_host release];
    
    [super dealloc];
}

- (NSString *)description
{
    return [NSString stringWithFormat:@"<%@[0x%lx] \"%@\": %dx%d, %dbpp, %d bit depth, %@%@colourspace: max: %d/%d/%d shift: %d/%d/%d>",
        [[self class] description], self,
        _desktopName,
        (int)_width, (int)_height, (int)_bitsPerPixel, (int)_depth,
        (_bigEndian ? @"big endian, " : @""), (_trueColour ? @"true colour, " : @""),
        (int)_redMax, (int)_greenMax, (int)_blueMax,
        (int)_redShift, (int)_greenShift, (int)_blueShift];
}

- (unsigned)read24BitColor
{
    switch (_bitsPerPixel) {
        case 32:
            return NSSwapBigIntToHost([_socket readUnsignedInt]);

        case 16:
            return RFBConvertPixel16ToPixel32([_socket readUnsignedShort]);
        case 8:
            return RFBConvertPixel8ToPixel32([_socket readByte]);

        default:
            [NSException raise:NSInvalidArgumentException
                        format:@"color depth exception: not 8, 16, or 32"];
    }

    return 0;
}

@end

