#import "VNC.h"

NSString *VNCDisplayWillTerminateNotification = @"VNCDisplayWillTerminateNotification";

@implementation VNCDisplayManager

- (id)initWithAdaptor:(RFBProtocolAdaptor *)adaptor
{
    NSPoint origin;

    if (self = [super init]) {
        NSRect contentRect = NSMakeRect(0, 0, [adaptor width], [adaptor height]);
        
        _adaptor = [adaptor retain];
        if (![NSBundle loadNibNamed:@"VNCDisplayManager" owner:self])
            [NSException raise:NSInternalInconsistencyException
                        format:@"couldn't load VNCDisplayManager.nib!"];

        [window setTitle:[_adaptor desktopName]];
        [window setContentSize:contentRect.size];
        if ([NSApp mainWindow]) {
            origin.x = [[NSApp mainWindow] frame].origin.x + 40;
            origin.y = [[NSApp mainWindow] frame].origin.y - 40;
        }
        else {
            origin.x = 100;
            origin.y = [[NSScreen mainScreen] frame].size.height - 100 - [_adaptor height];
        }
        [window setFrameOrigin:origin];

        inputView = [[[VNCInputView alloc] initWithFrame:contentRect
                                                 adaptor:_adaptor] autorelease];
        displayView = [[[VNCDisplayView alloc] initWithFrame:contentRect] autorelease];

        [[statusTextField retain] removeFromSuperview];
        [[window contentView] addSubview:displayView];
        [[window contentView] addSubview:statusTextField];
        [[window contentView] addSubview:inputView];
        [statusTextField release];

        [window setInitialFirstResponder:inputView];
        [inputView addTrackingRect:[inputView frame]
                             owner:inputView
                          userData:@"" assumeInside:NO];
       
        _dataPort = [[NSPort alloc] init];
        [_dataPort setDelegate:self];
        [[NSRunLoop currentRunLoop] addPort:_dataPort forMode:NSDefaultRunLoopMode];

        [self requestFullUpdate:nil];
        
        [NSThread detachNewThreadSelector:@selector(protocolMonitor:)
                                 toTarget:self
                               withObject:_dataPort];

        _updateTimer = [[NSTimer scheduledTimerWithTimeInterval:1.0
                                                         target:self
                                                       selector:@selector(timedUpdateHandler:)
                                                       userInfo:nil
                                                        repeats:YES] retain];
        
        [window makeKeyAndOrderFront:nil];
        [window makeFirstResponder:inputView];

        [statusTextField setStringValue:@"Waiting for update..."];
    }

    return self;
}

- (void)requestFullUpdate:(id)sender
{
    [_adaptor sendFramebufferUpdateRequestForRect:[displayView frame] incremental:NO];
}

- (RFBProtocolAdaptor *)adaptor
{
    return _adaptor;
}

- (BOOL)windowShouldClose:(id)sender
{
    return [self performClose];
}

- (void)windowWillMiniaturize:(NSNotification *)notification
{
    // Now this is just gravy.
    NSImage *image = [[[[displayView framebuffer] framebufferImage] copy] autorelease];

    [image setScalesWhenResized:YES];
    [image setSize:NSMakeSize(48, 48)];

    [window setMiniwindowImage:image];
}

- (BOOL)performClose
{
    if (NSRunAlertPanel(@"Confirm Close",
                        @"Really close \"%@\"?",
                        @"Yes", @"No", nil,
                        [window title]) == NSAlertDefaultReturn) {
        [self close];
        return YES;
    }
    return NO;
}

- (void)invalidate
{
    [_updateTimer invalidate];
    [[NSRunLoop currentRunLoop] removePort:_dataPort forMode:NSDefaultRunLoopMode];
    [_dataPort release];

    [_adaptor release];
    _adaptor = nil;
}

- (void)close
{
    NSNotification *notification = [NSNotification notificationWithName:VNCDisplayWillTerminateNotification object:self];

    [self invalidate];
    [window close];
    [_updateTimer release];

    [[NSNotificationQueue defaultQueue] enqueueNotification:notification
                                               postingStyle:NSPostNow];
}

- (void)protocolMonitor:(NSPort *)dataPort
{
    NSAutoreleasePool *threadPool = [[NSAutoreleasePool alloc] init];

    NS_DURING {
        while(_adaptor != nil) {
            NSAutoreleasePool *loopPool = [[NSAutoreleasePool alloc] init];
            RFBMessage *message = [RFBMessage messageWithAdaptor:_adaptor];

            if (message == nil)
                break;

            [dataPort sendData:[NSArchiver archivedDataWithRootObject:message]];
            [loopPool release];
        }
    }
    NS_HANDLER {
        _threadException = [localException retain];
        [dataPort sendData:[NSData data]];
    }
    NS_ENDHANDLER;

    [threadPool release];
    [NSThread exit];
}

- (void)processMessage:(RFBMessage *)message
{
    switch([message type]) {
        case RFBFramebufferUpdate: {
            NSEnumerator *rectangleEnumerator = [[(id)message rectangles] objectEnumerator];
            RFBRectangle *rectangle;

            while (rectangle = [rectangleEnumerator nextObject])
                [rectangle drawWithFramebuffer:[displayView framebuffer]];
            
            if ([window isVisible]) {
                [displayView lockFocus];
                [[displayView framebuffer] updateMinimum];
                [displayView unlockFocus];
                [window flushWindow];
            }
            [statusTextField setStringValue:[NSString stringWithFormat:@"Processed %d rectangles", [[(id)message rectangles] count]]];

            break;
        }
            
        case RFBBell:
            NSBeep();		// wow.
            break;
            
        case RFBSetColourMapEntries:
            break;

        case RFBServerCutText:
            [[NSPasteboard generalPasteboard] declareTypes:[NSArray arrayWithObject:NSStringPboardType] owner:nil];
            [[NSPasteboard generalPasteboard] setString:[(id)message text]
                                                forType:NSStringPboardType];
            break;

        default:
            [NSException raise:NSInvalidArgumentException
                        format:@"unknown message %@", [message description]];
    }
    
}

- (void)handlePortMessage:(NSPortMessage *)message
{
    NSData *data = [[message components] objectAtIndex:0];

    if ([data length] > 0 && _adaptor != nil) {
        RFBMessage *message = [NSUnarchiver unarchiveObjectWithData:data];
        [self processMessage:message];
    }
    else {
        TCShowModelessAlertPanel(@"Network Error",
                                 @"Exception while monitoring %@:\n%@",
                                 [_adaptor desktopName], [_threadException description]);
        [self invalidate];
    }
}

- (void)timedUpdateHandler:(NSTimer *)timer
{
    [_adaptor sendFramebufferUpdateRequestForRect:[displayView frame] incremental:YES];
}

- (NSData *)desktopTIFFRepresentation
{
    return [[[displayView framebuffer] framebufferImage]
	TIFFRepresentationUsingCompression:NSTIFFCompressionJPEG factor:75.0];
}

@end
